import random
import re
import string
import unittest
from unittest import mock

from kinto.core.errors import ERRORS
from kinto.core.testing import FormattedErrorMixin

from .support import BaseWebTest, get_user_headers

RE_ISO8601 = re.compile(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}\+00:00")


class PostgresWebTest(BaseWebTest):
    source_bucket = "/buckets/alice"
    source_collection = "/buckets/alice/collections/scid"
    destination_bucket = "/buckets/alice"
    destination_collection = "/buckets/alice/collections/dcid"

    def setUp(self):
        super(PostgresWebTest, self).setUp()
        # Patch calls to Autograph.
        patch = mock.patch("kinto_remote_settings.signer.backends.autograph.requests")
        self.addCleanup(patch.stop)
        self.mocked_autograph = patch.start()

        def fake_sign():
            fake_signature = "".join(random.sample(string.ascii_lowercase, 10))
            return [
                {
                    "signature": fake_signature,
                    "hash_algorithm": "",
                    "signature_encoding": "",
                    "x5u": "",
                    "ref": "",
                }
            ]

        self.mocked_autograph.post.return_value.json.side_effect = fake_sign

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        settings["storage_backend"] = "kinto.core.storage.postgresql"
        db = "postgresql://postgres:postgres@localhost/testdb"
        settings["storage_url"] = db
        settings["permission_backend"] = "kinto.core.permission.postgresql"
        settings["permission_url"] = db
        settings["cache_backend"] = "kinto.core.cache.memory"

        settings["statsd_url"] = "udp://127.0.0.1:8125"

        settings["kinto.signer.resources"] = "%s -> %s" % (
            cls.source_collection,
            cls.destination_collection,
        )
        return settings


class SignoffWebTest(PostgresWebTest):
    def setUp(self):
        super(SignoffWebTest, self).setUp()
        self.headers = get_user_headers("tarte:en-pion")
        resp = self.app.get("/", headers=self.headers)
        self.userid = resp.json["user"]["id"]

        self.other_headers = get_user_headers("Sam:Wan Heilss")
        resp = self.app.get("/", headers=self.other_headers)
        self.other_userid = resp.json["user"]["id"]

        # Source bucket
        self.app.put_json(
            self.source_bucket,
            {"permissions": {"write": ["system.Authenticated"]}},
            headers=self.headers,
        )

        # Editors and reviewers group
        self.app.put_json(
            self.source_bucket + "/groups/editors",
            {"data": {"members": [self.userid, self.other_userid]}},
            headers=self.headers,
        )
        self.app.put_json(
            self.source_bucket + "/groups/reviewers",
            {"data": {"members": [self.userid, self.other_userid]}},
            headers=self.headers,
        )

        # Source collection with 2 records
        self.app.put_json(self.source_collection, headers=self.headers)
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "hello"}},
            headers=self.headers,
        )
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"id": "in-french", "title": "bonjour"}},
            headers=self.headers,
        )


class CollectionStatusTest(SignoffWebTest, FormattedErrorMixin, unittest.TestCase):
    def test_status_can_be_refreshed_even_if_never_signed(self):
        resp = self.app.get(self.source_collection, headers=self.headers)
        before = resp.json["data"]["last_signature_date"]

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)

        assert resp.json["data"]["status"] == "work-in-progress"
        after = resp.json["data"]["last_signature_date"]
        assert before < after
        # The review request / approval field are not set.
        assert "last_review_date" not in resp.json["data"]
        assert "last_review_request_date" not in resp.json["data"]

    def test_status_cannot_be_set_to_unknown_value(self):
        resp = self.app.patch_json(
            self.source_collection,
            {"data": {"status": "married"}},
            headers=self.headers,
            status=400,
        )
        self.assertFormattedError(
            response=resp,
            code=400,
            errno=ERRORS.INVALID_POSTED_DATA,
            error="Bad Request",
            message="Invalid status 'married'",
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "work-in-progress"

    def test_status_cannot_be_set_to_signed_manually(self):
        resp = self.app.patch_json(
            self.source_collection,
            {"data": {"status": "signed"}},
            headers=self.headers,
            status=400,
        )
        assert resp.json["message"] == "Cannot set status to 'signed'"
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "work-in-progress"

    def test_status_can_be_set_to_work_in_progress_manually(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "work-in-progress"}},
            headers=self.headers,
        )

    def test_status_can_be_maintained_as_signed_manually(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        # Signature occured, the source collection will be signed.
        self.app.patch_json(
            self.source_collection, {"data": {"status": "signed"}}, headers=self.headers
        )
        self.app.patch_json(
            self.source_collection, {"data": {"author": "dali"}}, headers=self.headers
        )

    def test_status_cannot_be_removed_once_it_was_set(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        self.app.put_json(
            self.source_collection, {"data": {}}, headers=self.headers, status=400
        )

    def test_status_cannot_be_emptied_once_it_was_set(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        self.app.put_json(
            self.source_collection,
            {"data": {"status": ""}},
            headers=self.headers,
            status=400,
        )

    def test_status_can_be_reset(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "to-review"

    def test_status_is_set_to_work_in_progress_when_records_are_posted(self):
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "work-in-progress"

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "Hallo"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "work-in-progress"

    def test_statsd_reports_number_of_records_approved(self):
        # Publish the changes to reset the setup.
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        # Make some changes.
        self.app.post(self.source_collection + "/records", headers=self.headers)
        self.app.delete(
            self.source_collection + "/records/in-french", headers=self.headers
        )

        # Publish the changes again.
        statsd_client = self.app.app.registry.statsd
        with mock.patch.object(statsd_client, "count") as mocked:
            self.app.patch_json(
                self.source_collection,
                {"data": {"status": "to-sign"}},
                headers=self.headers,
            )
        call_args = mocked.call_args_list[0][0]

        # One creation and one deletion.
        assert call_args == ("plugins.signer.approved_changes", 2)


class ForceReviewTest(SignoffWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["signer.to_review_enabled"] = "true"
        return settings

    def test_status_cannot_be_set_to_to_sign_without_review(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
            status=400,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "work-in-progress"

    def test_passing_from_signed_to_to_sign_is_not_allowed(self):
        """This is useful when the x5u certificate changed and you want
        to retrigger a new signature."""
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )
        resp = self.app.get(self.source_collection, headers=self.other_headers)
        assert resp.json["data"]["status"] == "signed"

        resp = self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
            status=400,
        )
        assert resp.json["message"] == "Collection already signed"

    def test_editor_cannot_be_reviewer(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
            status=403,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "to-review"

        # Try again as someone else
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"


class RefreshSignatureTest(SignoffWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["signer.to_review_enabled"] = "true"
        return settings

    def setUp(self):
        super().setUp()
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

    def test_resign_does_not_bump_destination_timestamp(self):
        resp = self.app.head(
            self.destination_collection + "/records", headers=self.headers
        )
        before = resp.headers["ETag"]

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=self.headers,
        )

        resp = self.app.head(
            self.destination_collection + "/records", headers=self.headers
        )
        after = resp.headers["ETag"]
        assert before == after

    def test_request_review_and_re_sign_does_not_bump_destination_timestamp(self):
        resp = self.app.head(
            self.destination_collection + "/records", headers=self.headers
        )
        before = resp.headers["ETag"]

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.other_headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )

        resp = self.app.head(
            self.destination_collection + "/records", headers=self.headers
        )
        after = resp.headers["ETag"]
        assert before == after

    def test_editor_can_retrigger_a_signature(self):
        # Editor retriggers a signature, without going through review.
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=self.headers,
        )

        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

    def test_reviewer_can_retrigger_a_signature(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=self.other_headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

    def test_if_resign_fails_signature_is_rolledback(self):
        self.mocked_autograph.post.side_effect = ValueError("Boom!")

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=self.other_headers,
            status=500,
        )

        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

    def test_non_reviewer_can_retrigger_a_signature(self):
        writer_headers = get_user_headers("walter:white")
        resp = self.app.get("/", headers=writer_headers)
        writer_userid = resp.json["user"]["id"]
        self.app.patch_json(
            self.source_bucket,
            {"permissions": {"write": [writer_userid]}},
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=writer_headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

    def test_signature_can_be_refreshed_with_pending_changes(self):
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "pending change"}},
            headers=self.headers,
        )

        resp = self.app.get(self.destination_collection, headers=self.headers)
        before_signature = resp.json["data"]["signature"]["signature"]

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "work-in-progress"

        resp = self.app.get(self.destination_collection, headers=self.headers)
        assert resp.json["data"]["signature"]["signature"] != before_signature


class TrackingFieldsTest(SignoffWebTest, unittest.TestCase):
    def last_edit_by_and_date_are_tracked(self):
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "Hallo"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["last_edit_by"] == self.userid
        assert RE_ISO8601.match(resp.json["data"]["last_edit_date"])

    def test_last_review_request_by_and_date_are_tracked(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["last_review_request_by"] == self.userid
        assert RE_ISO8601.match(resp.json["data"]["last_review_request_date"])

    def test_last_review_by_and_date_are_tracked(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"
        assert resp.json["data"]["last_review_by"] == self.userid
        assert RE_ISO8601.match(resp.json["data"]["last_review_date"])
        assert resp.json["data"]["last_signature_by"] == self.userid
        assert RE_ISO8601.match(resp.json["data"]["last_signature_date"])

    def test_last_review_differs_from_last_signature_on_refresh_signature(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"
        last_reviewer = resp.json["data"]["last_review_by"]

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=self.headers,
        )
        metadata = self.app.get(self.source_collection, headers=self.headers).json[
            "data"
        ]
        assert metadata["status"] == "signed"

        assert metadata["last_signature_date"] != metadata["last_review_date"]
        assert last_reviewer == metadata["last_review_by"]

    def test_editor_can_be_reviewer(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

    def test_editor_reviewer_editor_cannot_be_changed_nor_removed(self):
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "Hallo"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )

        resp = self.app.get(self.source_collection, headers=self.headers)
        source_collection = resp.json["data"]
        assert source_collection["status"] == "signed"

        # All tracking fields are here.
        expected = (
            "last_edit_by",
            "last_edit_date",
            "last_review_request_by",
            "last_review_request_date",
            "last_review_by",
            "last_review_date",
            "last_signature_by",
            "last_signature_date",
        )
        assert all([f in source_collection for f in expected])

        # They cannot be changed nor removed.
        for f in expected:
            self.app.patch_json(
                self.source_collection,
                {"data": {f: "changed"}},
                headers=self.headers,
                status=400,
            )
            changed = source_collection.copy()
            changed.pop(f)
            self.app.put_json(
                self.source_collection,
                {"data": changed},
                headers=self.headers,
                status=400,
            )


class RollbackChangesTest(SignoffWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        cls.preview_bucket = "/buckets/preview"
        cls.preview_collection = cls.preview_bucket + "/collections/pcid"

        settings["signer.to_review_enabled"] = "true"
        settings["kinto.signer.resources"] = " -> ".join(
            [cls.source_collection, cls.preview_collection, cls.destination_collection]
        )
        return settings

    def setUp(self):
        super().setUp()
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

        self.app.put_json(
            self.source_collection + "/records/r1",
            {"data": {"title": "Hallo"}},
            headers=self.headers,
        )
        self.app.put_json(
            self.source_collection + "/records/r2",
            {"data": {"title": "Bon dia"}},
            headers=self.headers,
        )
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "work-in-progress"

    def test_return_400_if_status_is_signed(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )

        resp = self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
            status=400,
        )

        assert resp.json["message"] == "Collection has no work-in-progress"

    def test_rollbacks_if_no_pending_changes(self):
        self.app.delete(self.source_collection + "/records/r1", headers=self.headers)
        self.app.delete(self.source_collection + "/records/r2", headers=self.headers)

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )

        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

    def test_rollbacks_if_review_already_requested(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )

        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

    def test_tracking_fields_are_updated(self):
        resp = self.app.get(self.source_collection, headers=self.headers)
        before_date = resp.json["data"]["last_edit_date"]

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )

        resp = self.app.get(self.source_collection, headers=self.headers)
        after_date = resp.json["data"]["last_edit_date"]
        assert before_date != after_date

    def test_comments_are_reset(self):
        self.app.patch_json(
            self.source_collection,
            {
                "data": {
                    "last_editor_comment": "please check that",
                    "last_reviewer_comment": "looks good",
                }
            },
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )

        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["last_editor_comment"] == ""
        assert resp.json["data"]["last_reviewer_comment"] == ""

    def test_recreates_deleted_record(self):
        resp = self.app.delete(
            self.source_collection + "/records?_limit=1&_sort=last_modified",
            headers=self.headers,
        )
        deleted_id = resp.json["data"][0]["id"]

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )

        resp = self.app.get(
            self.source_collection + f"/records/{deleted_id}",
            headers=self.headers,
            status=200,
        )
        assert resp.json["data"]["title"] == "hello"

    def test_reverts_updated_records(self):
        resp = self.app.get(
            self.source_collection + "/records?_limit=1&_sort=last_modified",
            headers=self.headers,
        )
        update_id = resp.json["data"][0]["id"]
        self.app.put_json(
            self.source_collection + f"/records/{update_id}",
            {"data": {"title": "Ave"}},
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )

        resp = self.app.get(
            self.source_collection + f"/records/{update_id}",
            headers=self.headers,
            status=200,
        )
        assert resp.json["data"]["title"] == "hello"

    def test_removes_created_records(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )

        self.app.get(
            self.source_collection + "/records/r1", headers=self.headers, status=404
        )
        self.app.get(
            self.source_collection + "/records/r2", headers=self.headers, status=404
        )

    def test_also_resets_changes_on_preview(self):
        resp = self.app.get(self.preview_collection + "/records", headers=self.headers)
        size_setup = len(resp.json["data"])
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        resp = self.app.get(self.preview_collection + "/records", headers=self.headers)
        size_before = len(resp.json["data"])
        assert size_setup != size_before

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )

        resp = self.app.get(self.preview_collection + "/records", headers=self.headers)
        size_after = len(resp.json["data"])
        assert size_before != size_after
        assert size_setup == size_after

    def test_preview_signature_is_refreshed(self):
        resp = self.app.get(self.preview_collection, headers=self.headers)
        sign_before = resp.json["data"]["signature"]["signature"]

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )

        resp = self.app.get(self.preview_collection, headers=self.headers)
        sign_after = resp.json["data"]["signature"]["signature"]
        assert sign_before != sign_after

    def test_does_not_recreate_tombstones(self):
        # Approve creation of r1 and r2.
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )
        # Delete r1.
        self.app.delete(
            self.source_collection + "/records/r1",
            headers=self.headers,
        )
        # Approve deletion of r1.
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )
        # Recreate r1.
        self.app.put_json(
            self.source_collection + "/records/r1",
            {"data": {"title": "Servus"}},
            headers=self.headers,
        )
        # Rollback.
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )
        # r1 should be deleted again.
        self.app.get(
            self.source_collection + "/records/r1", headers=self.headers, status=404
        )
        self.app.get(
            self.preview_collection + "/records/r1", headers=self.headers, status=404
        )
        self.app.get(
            self.destination_collection + "/records/r1",
            headers=self.headers,
            status=404,
        )


class UserGroupsTest(SignoffWebTest, FormattedErrorMixin, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["signer.to_review_enabled"] = "true"
        return settings

    def setUp(self):
        super(UserGroupsTest, self).setUp()
        self.editor_headers = get_user_headers("edith:her")
        resp = self.app.get("/", headers=self.editor_headers)
        self.editor = resp.json["user"]["id"]

        self.editor_headers = get_user_headers("emo:billier")
        resp = self.app.get("/", headers=self.editor_headers)
        self.editor = resp.json["user"]["id"]

        self.reviewer_headers = get_user_headers("ray:weaver")
        resp = self.app.get("/", headers=self.reviewer_headers)
        self.reviewer = resp.json["user"]["id"]

        self.app.put_json(
            "/buckets/alice/groups/editors",
            {"data": {"members": [self.editor]}},
            headers=self.headers,
        )

        self.app.put_json(
            "/buckets/alice/groups/reviewers",
            {"data": {"members": [self.reviewer]}},
            headers=self.headers,
        )

    def test_only_editors_can_ask_to_review(self):
        resp = self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.reviewer_headers,
            status=403,
        )
        self.assertFormattedError(
            response=resp,
            code=403,
            errno=ERRORS.FORBIDDEN,
            error="Forbidden",
            message="Not in editors group",
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.editor_headers,
        )

    def test_only_reviewers_can_ask_to_sign(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.editor_headers,
        )

        resp = self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.editor_headers,
            status=403,
        )
        self.assertFormattedError(
            response=resp,
            code=403,
            errno=ERRORS.FORBIDDEN,
            error="Forbidden",
            message="Not in reviewers group",
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.reviewer_headers,
        )


class SpecificUserGroupsTest(SignoffWebTest, FormattedErrorMixin, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        cls.source_collection1 = "/buckets/alice/collections/cid1"
        cls.source_collection2 = "/buckets/alice/collections/cid2"

        settings["kinto.signer.resources"] = "%s -> %s\n%s -> %s" % (
            cls.source_collection1,
            cls.source_collection1.replace("alice", "destination"),
            cls.source_collection2,
            cls.source_collection2.replace("alice", "destination"),
        )

        settings["signer.to_review_enabled"] = "true"

        settings["signer.alice.cid1.editors_group"] = "editeurs"
        settings["signer.alice.cid1.reviewers_group"] = "revoyeurs"
        return settings

    def setUp(self):
        super(SpecificUserGroupsTest, self).setUp()

        self.app.put_json(self.source_collection1, headers=self.headers)
        self.app.put_json(self.source_collection2, headers=self.headers)

        self.someone_headers = get_user_headers("sam:wan")

        self.editor_headers = get_user_headers("emo:billier")
        resp = self.app.get("/", headers=self.editor_headers)
        self.editor = resp.json["user"]["id"]

        self.app.put_json(
            "/buckets/alice/groups/editeurs",
            {"data": {"members": [self.editor]}},
            headers=self.headers,
        )

    def test_editors_cannot_ask_to_review_if_not_specifically_configured(self):
        resp = self.app.patch_json(
            self.source_collection2,
            {"data": {"status": "to-review"}},
            headers=self.someone_headers,
            status=403,
        )
        self.assertFormattedError(
            response=resp,
            code=403,
            errno=ERRORS.FORBIDDEN,
            error="Forbidden",
            message="Not in editors group",
        )

    def test_only_specific_editors_can_ask_to_review(self):
        resp = self.app.patch_json(
            self.source_collection1,
            {"data": {"status": "to-review"}},
            headers=self.someone_headers,
            status=403,
        )
        self.assertFormattedError(
            response=resp,
            code=403,
            errno=ERRORS.FORBIDDEN,
            error="Forbidden",
            message="Not in editeurs group",
        )

    def test_only_reviewers_can_ask_to_sign(self):
        self.app.patch_json(
            self.source_collection1,
            {"data": {"status": "to-review"}},
            headers=self.editor_headers,
        )
        resp = self.app.patch_json(
            self.source_collection1,
            {"data": {"status": "to-sign"}},
            headers=self.editor_headers,
            status=403,
        )
        self.assertFormattedError(
            response=resp,
            code=403,
            errno=ERRORS.FORBIDDEN,
            error="Forbidden",
            message="Not in revoyeurs group",
        )


class PreviewCollectionTest(SignoffWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        cls.preview_bucket = "/buckets/preview"
        cls.preview_collection = cls.preview_bucket + "/collections/pcid"

        settings["signer.to_review_enabled"] = "true"
        settings["kinto.signer.resources"] = "%s -> %s -> %s" % (
            cls.source_collection,
            cls.preview_collection,
            cls.destination_collection,
        )
        return settings

    def test_the_preview_collection_is_updated_and_signed(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )

        self.app.get(self.preview_bucket, headers=self.headers)

        resp = self.app.get(self.preview_collection, headers=self.headers)
        assert "signature" in resp.json["data"]

        resp = self.app.get(self.preview_collection + "/records", headers=self.headers)
        assert len(resp.json["data"]) == 2

    def test_the_preview_collection_receives_kinto_admin_ui_attributes(self):
        self.app.patch_json(
            self.source_collection,
            {
                "data": {
                    "status": "to-review",
                    "displayFields": ["age"],
                    "schema": {"type": "object"},
                }
            },
            headers=self.headers,
        )

        resp = self.app.get(self.preview_collection, headers=self.headers)
        assert resp.json["data"]["displayFields"] == ["age"]
        assert "schema" in resp.json["data"]

    def test_the_preview_collection_is_also_resigned(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        resp = self.app.get(self.preview_collection, headers=self.headers)
        signature_preview_before = resp.json["data"]["signature"]
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )
        resp = self.app.get(self.destination_collection, headers=self.headers)
        signature_destination_before = resp.json["data"]["signature"]
        # status is signed.
        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

        # Resign.
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=self.headers,
        )

        resp = self.app.get(self.destination_collection, headers=self.headers)
        signature_destination_after = resp.json["data"]["signature"]
        assert signature_destination_before != signature_destination_after
        resp = self.app.get(self.preview_collection, headers=self.headers)
        signature_preview_after = resp.json["data"]["signature"]
        assert signature_preview_before != signature_preview_after

    def test_the_preview_collection_is_emptied_when_source_records_are_deleted(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )

        resp = self.app.get(self.source_collection + "/records", headers=self.headers)
        records = resp.json["data"]
        for r in records:
            self.app.delete(
                self.source_collection + "/records/" + r["id"], headers=self.headers
            )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )

        resp = self.app.get(self.preview_collection + "/records", headers=self.headers)
        records = resp.json["data"]
        assert len(records) == 0

    def test_the_preview_collection_is_emptied_when_source_is_deleted(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )

        self.app.delete(self.source_collection + "/records", headers=self.headers).json[
            "data"
        ]
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )

        resp = self.app.get(self.preview_collection + "/records", headers=self.headers)
        records = resp.json["data"]
        assert len(records) == 0

    def test_last_editor_comment_are_reset_on_review(self):
        self.app.patch_json(
            self.source_collection,
            {
                "data": {
                    "last_editor_comment": "please check that",
                    "last_reviewer_comment": "looks good",
                }
            },
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )

        resp = self.app.get(self.source_collection, headers=self.headers)
        assert resp.json["data"]["last_editor_comment"] == ""


class CollectionDelete(SignoffWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        cls.source_bucket = "/buckets/source"
        cls.source_collection = cls.source_bucket + "/collections/cid"
        cls.preview_bucket = "/buckets/preview"
        cls.preview_collection = cls.preview_bucket + "/collections/cid"
        cls.destination_bucket = "/buckets/destination"
        cls.destination_collection = cls.destination_bucket + "/collections/cid"

        settings["signer.to_review_enabled"] = "true"
        settings["kinto.signer.resources"] = (
            "%s -> %s -> %s"
            % (cls.source_bucket, cls.preview_bucket, cls.destination_bucket)
            + "\n %s -> %s"
            % (
                cls.source_bucket + "/collections/no-preview",
                cls.destination_bucket + "/collections/no-preview",
            )
            + "\n /buckets/some-bucket -> /buckets/some-other"
        )
        return settings

    def setUp(self):
        super(CollectionDelete, self).setUp()

        self.app.put(
            self.source_bucket + "/collections/no-preview", headers=self.headers
        )
        self.app.put(
            self.destination_bucket + "/collections/extra", headers=self.headers
        )
        self.app.put(
            self.preview_bucket + "/collections/no-preview", headers=self.headers
        )

    def test_cannot_delete_preview_collection_if_used(self):
        self.app.delete(self.preview_collection, headers=self.headers, status=403)

    def test_cannot_delete_destination_collection_if_used(self):
        self.app.delete(self.destination_collection, headers=self.headers, status=403)
        self.app.delete(
            self.destination_bucket + "/collections/no-preview",
            headers=self.headers,
            status=403,
        )

    def test_can_delete_preview_if_source_is_deleted(self):
        self.app.delete(self.source_collection, headers=self.headers)
        self.app.delete(self.preview_collection, headers=self.headers)

    def test_can_delete_preview_if_unused(self):
        self.app.delete(
            self.preview_bucket + "/collections/no-preview", headers=self.headers
        )

    def test_can_delete_destination_if_unused(self):
        self.app.delete(
            self.destination_bucket + "/collections/extra", headers=self.headers
        )


class NoReviewTest(SignoffWebTest, unittest.TestCase):
    """
    If preview collection is set in config, we create it
    and copy the records there, even if review is disabled.
    """

    source_bucket = "/buckets/dev"
    source_collection = "/buckets/dev/collections/normandy"
    preview_bucket = "/buckets/stage"
    preview_collection = "/buckets/stage/collections/normandy"
    destination_bucket = "/buckets/prod"
    destination_collection = "/buckets/prod/collections/normandy"

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        # preview collection exists.
        settings["kinto.signer.resources"] = " -> ".join(
            (cls.source_bucket, cls.preview_bucket, cls.destination_bucket)
        )
        # dev/onecrl has review enabled.
        settings["signer.to_review_enabled"] = "true"
        # dev/normandy has review disabled.
        settings["signer.dev.normandy.to_review_enabled"] = "false"

        return settings

    def setUp(self):
        super().setUp()
        # Make the preview bucket readable (to obtain explicit 404 when collections
        # don't exist instead of ambiguous 403)
        self.app.put_json(
            self.preview_bucket,
            {"permissions": {"read": ["system.Everyone"]}},
            headers=self.headers,
        )
        self.app.put(self.source_bucket + "/collections/onecrl", headers=self.headers)
        self.app.put(self.source_collection, headers=self.headers)

    def test_the_preview_collection_is_created_when_review_enabled(self):
        self.app.get(self.preview_bucket + "/collections/onecrl", headers=self.headers)

    def test_the_preview_collection_is_created_when_review_disabled(self):
        self.app.get(self.preview_collection, headers=self.headers)

    def test_the_preview_collection_is_updated_when_review_enabled(self):
        before = len(
            self.app.get(
                self.preview_bucket + "/collections/onecrl/records",
                headers=self.headers,
            ).json["data"]
        )
        self.app.post_json(
            self.source_bucket + "/collections/onecrl/records",
            {"data": {"title": "Hallo"}},
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_bucket + "/collections/onecrl",
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )

        after = len(
            self.app.get(
                self.preview_bucket + "/collections/onecrl/records",
                headers=self.headers,
            ).json["data"]
        )
        assert after > before, "Preview was not updated when review enabled"

    def test_the_preview_collection_is_updated_when_review_disabled(self):
        before = len(
            self.app.get(
                self.preview_collection + "/records", headers=self.headers
            ).json["data"]
        )
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "Hallo"}},
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )

        after = len(
            self.app.get(
                self.preview_collection + "/records", headers=self.headers
            ).json["data"]
        )
        assert after > before, "Preview was not updated when review disabled"


class NoPreviewTest(SignoffWebTest, unittest.TestCase):
    """
    If preview collection is not set in config, we don't create it
    even if review is enabled.
    """

    source_bucket = "/buckets/dev"
    source_collection = "/buckets/dev/collections/normandy"
    preview_bucket = "/buckets/preview"
    destination_bucket = "/buckets/prod"
    destination_collection = "/buckets/prod/collections/normandy"

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        # No preview collection.
        settings["kinto.signer.resources"] = " -> ".join(
            (cls.source_bucket, cls.destination_bucket)
        )
        # dev/onecrl has review enabled.
        settings["signer.to_review_enabled"] = "true"
        # dev/normandy has review disabled.
        settings["signer.dev.normandy.to_review_enabled"] = "false"

        return settings

    def setUp(self):
        super().setUp()
        self.app.put(self.preview_bucket, headers=self.headers)
        self.app.put(self.source_bucket + "/collections/onecrl", headers=self.headers)
        self.app.put_json(self.source_collection, headers=self.headers)

    def test_the_preview_collection_is_not_created_when_review_enabled(self):
        self.app.get(
            self.preview_bucket + "/collections/onecrl",
            status=404,
            headers=self.headers,
        )

    def test_the_preview_collection_is_not_created_when_review_disabled(self):
        self.app.get(
            self.preview_bucket + "/collections/normandy",
            status=404,
            headers=self.headers,
        )

    def test_the_preview_collection_is_not_updated_when_review_enabled(self):
        self.app.post_json(
            self.source_bucket + "/collections/onecrl/records",
            {"data": {"title": "Hallo"}},
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_bucket + "/collections/onecrl",
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )

        # The preview still doesn't exist
        self.app.get(
            self.preview_bucket + "/collections/onecrl",
            status=404,
            headers=self.headers,
        )
        # Destination will updated on review approval.

    def test_the_preview_collection_is_not_updated_when_review_disabled(self):
        before = len(
            self.app.get(
                self.destination_collection + "/records", headers=self.headers
            ).json["data"]
        )
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "Hallo"}},
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )

        # The preview still doesn't exist
        self.app.get(
            self.preview_bucket + "/collections/normandy",
            status=404,
            headers=self.headers,
        )
        # And destination was updated.
        after = len(
            self.app.get(
                self.destination_collection + "/records", headers=self.headers
            ).json["data"]
        )
        assert after > before, "Destination was not updated when review enabled"


class PerBucketTest(SignoffWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        cls.source_bucket = "/buckets/stage"
        cls.source_collection = cls.source_bucket + "/collections/cid"
        cls.preview_bucket = "/buckets/preview"
        cls.preview_collection = cls.preview_bucket + "/collections/cid"
        cls.destination_bucket = "/buckets/prod"
        cls.destination_collection = cls.destination_bucket + "/collections/cid"

        settings["kinto.signer.resources"] = " -> ".join(
            [cls.source_bucket, cls.preview_bucket, cls.destination_bucket]
        )

        settings["signer.to_review_enabled"] = "true"
        settings["signer.stage.specific.to_review_enabled"] = "false"

        settings["signer.stage.specific.autograph.hawk_id"] = "for-specific"
        return settings

    def test_destination_and_preview_collections_are_created_and_signed(self):
        col_uri = "/collections/pim"
        self.app.put(self.source_bucket + col_uri, headers=self.headers)

        data = self.app.get(self.preview_bucket + col_uri, headers=self.headers).json[
            "data"
        ]
        assert "signature" in data

        data = self.app.get(
            self.destination_bucket + col_uri, headers=self.headers
        ).json["data"]
        assert "signature" in data

        # Source status was set to signed.
        data = self.app.get(self.source_bucket + col_uri, headers=self.headers).json[
            "data"
        ]
        assert data["status"] == "signed"

    def test_review_settings_can_be_overriden_for_a_specific_collection(self):
        # review is not enabled for this particular one, sign directly!
        self.app.put_json(
            self.source_bucket + "/collections/specific",
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )

    def test_signer_can_be_specified_per_collection(self):
        self.mocked_autograph.post.reset_mock()
        self.app.put_json(
            self.source_bucket + "/collections/specific",
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )

        args, kwargs = self.mocked_autograph.post.call_args_list[0]
        assert args[0].startswith("http://localhost:8000")  # global.
        assert kwargs["auth"].credentials["id"] == "for-specific"
        assert (
            kwargs["auth"].credentials["key"].startswith("fs5w")
        )  # global in signer.ini


class GroupCreationTest(PostgresWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        cls.source_bucket = "/buckets/stage"
        cls.preview_bucket = "/buckets/preview"
        cls.destination_bucket = "/buckets/prod"

        settings["signer.to_review_enabled"] = "true"

        settings["kinto.signer.editors_group"] = "best-editors"
        settings["kinto.signer.reviewers_group"] = "{collection_id}-reviewers"
        settings["kinto.signer.resources"] = ";".join(
            [cls.source_bucket, cls.preview_bucket, cls.destination_bucket]
        )

        cls.editors_group = cls.source_bucket + "/groups/best-editors"
        cls.reviewers_group = cls.source_bucket + "/groups/good-reviewers"
        cls.source_collection = cls.source_bucket + "/collections/good"

        return settings

    def setUp(self):
        super(GroupCreationTest, self).setUp()

        resp = self.app.get("/", headers=self.headers)
        self.userid = resp.json["user"]["id"]

        self.app.put(self.source_bucket, headers=self.headers)

        self.other_headers = get_user_headers("otra:persona")
        resp = self.app.get("/", headers=self.other_headers)
        self.other_userid = resp.json["user"]["id"]

    def test_groups_are_not_touched_if_existing(self):
        resp = self.app.put(self.editors_group, headers=self.headers)
        before = resp.json["data"]["last_modified"]

        self.app.put(self.source_collection, headers=self.headers)

        resp = self.app.get(self.editors_group, headers=self.headers)
        after = resp.json["data"]["last_modified"]

        assert before == after

    def test_groups_are_created_if_missing(self):
        self.app.get(self.editors_group, headers=self.headers, status=404)
        self.app.get(self.reviewers_group, headers=self.headers, status=404)

        self.app.put(self.source_collection, headers=self.headers)

        self.app.get(self.editors_group, headers=self.headers)
        self.app.get(self.reviewers_group, headers=self.headers)

    def test_groups_are_allowed_to_write_the_source_collection(self):
        body = {"data": {"members": [self.other_userid]}}
        self.app.put_json(self.editors_group, body, headers=self.headers)

        self.app.put(self.source_collection, headers=self.headers)

        self.app.post_json(
            self.source_collection + "/records", headers=self.other_headers, status=201
        )

    def test_events_are_sent(self):
        patch = mock.patch("kinto_remote_settings.signer.utils.notify_resource_event")
        mocked = patch.start()
        self.addCleanup(patch.stop)

        self.app.put(self.source_collection, headers=self.headers)

        args, kwargs = mocked.call_args_list[0]
        _, fakerequest = args
        assert fakerequest["method"] == "PUT"
        assert fakerequest["path"] == "/buckets/stage/groups/best-editors"
        assert kwargs["resource_name"] == "group"

    def test_groups_permissions_include_current_user_only(self):
        self.app.put(self.source_collection, headers=self.headers)

        r = self.app.get(self.editors_group, headers=self.headers).json
        assert r["permissions"]["write"] == [self.userid]
        r = self.app.get(self.reviewers_group, headers=self.headers).json
        assert r["permissions"]["write"] == [self.userid]

    def test_editors_contains_current_user_as_member_by_default(self):
        self.app.put(self.source_collection, headers=self.headers)

        r = self.app.get(self.editors_group, headers=self.headers).json
        assert r["data"]["members"] == [self.userid]
        r = self.app.get(self.reviewers_group, headers=self.headers).json
        assert r["data"]["members"] == []

    def test_groups_are_not_touched_if_already_exist(self):
        resp = self.app.put(self.editors_group, headers=self.headers)
        editors_timetamp = resp.json["data"]["last_modified"]
        resp = self.app.put(self.reviewers_group, headers=self.headers)
        reviewers_timetamp = resp.json["data"]["last_modified"]

        self.app.put(self.source_collection, headers=self.headers)

        r = self.app.get(self.editors_group, headers=self.headers).json
        assert r["data"]["last_modified"] == editors_timetamp
        r = self.app.get(self.reviewers_group, headers=self.headers).json
        assert r["data"]["last_modified"] == reviewers_timetamp

    def test_groups_are_not_created_if_not_allowed(self):
        # Allow this other user to create collections.
        body = {"permissions": {"collection:create": [self.other_userid]}}
        self.app.patch_json(self.source_bucket, body, headers=self.headers)

        # Create the collection.
        self.app.put(self.source_collection, headers=self.other_headers)

        # Groups were not created.
        self.app.get(self.editors_group, headers=self.headers, status=404)
        self.app.get(self.reviewers_group, headers=self.headers, status=404)

    def test_groups_are_created_if_allowed_via_group_create_perm(self):
        # Allow this other user to create collections and groups.
        body = {
            "permissions": {
                "collection:create": [self.other_userid],
                "group:create": [self.other_userid],
            }
        }
        self.app.patch_json(self.source_bucket, body, headers=self.headers)

        # Create the collection.
        self.app.put(self.source_collection, headers=self.other_headers)

        self.app.get(self.editors_group, headers=self.headers)
        self.app.get(self.reviewers_group, headers=self.headers)



import sys

from .backends.local_ecdsa import ECDSASigner


def generate_keypair(private_key_location, public_key_location):
    private_key, public_key = ECDSASigner.generate_keypair()

    with open(private_key_location, "wb+") as tmp_file:
        tmp_file.write(private_key)

    with open(public_key_location, "wb+") as tmp_file:
        tmp_file.write(public_key)


if __name__ == "__main__":  # pragma: no cover
    if len(sys.argv) != 3:
        print("Usage: python -m signer.generate_keypair " "{public_key} {private_key}")
        sys.exit(0)
    generate_keypair(sys.argv[1], sys.argv[2])



class BaseEvent(object):
    def __init__(self, request, payload, impacted_objects, resource, original_event):
        self.request = request
        self.payload = payload
        self.impacted_objects = impacted_objects
        self.resource = resource
        self.original_event = original_event

    @property
    def impacted_records(self):
        return self.impacted_objects


class ReviewRequested(BaseEvent):
    def __init__(self, changes_count, comment, **kwargs):
        super().__init__(**kwargs)
        self.comment = comment
        self.changes_count = changes_count
        self.payload["comment"] = comment
        self.payload["changes_count"] = changes_count


class ReviewRejected(BaseEvent):
    def __init__(self, comment, **kwargs):
        super().__init__(**kwargs)
        self.comment = comment
        self.payload["comment"] = comment


class ReviewApproved(BaseEvent):
    def __init__(self, changes_count, **kwargs):
        super().__init__(**kwargs)
        self.changes_count = changes_count
        self.payload["changes_count"] = changes_count


class ReviewCanceled(BaseEvent):
    def __init__(self, changes_count, **kwargs):
        super().__init__(**kwargs)
        self.changes_count = changes_count
        self.payload["changes_count"] = changes_count



import datetime
import logging
from enum import Enum

from kinto.core.events import ACTIONS
from kinto.core.storage.exceptions import RecordNotFoundError
from pyramid.security import Everyone

from .serializer import canonical_json
from .utils import STATUS, ensure_resource_exists, notify_resource_event, records_diff

logger = logging.getLogger(__name__)


FIELD_ID = "id"
FIELD_LAST_MODIFIED = "last_modified"
# Source collection fields to be copied to destination.
PUBLISHED_COLLECTION_FIELDS = ("schema", "sort", "displayFields", "attachment")


class TRACKING_FIELDS(Enum):
    LAST_EDIT_BY = "last_edit_by"
    LAST_EDIT_DATE = "last_edit_date"
    LAST_REVIEW_REQUEST_BY = "last_review_request_by"
    LAST_REVIEW_REQUEST_DATE = "last_review_request_date"
    LAST_REVIEW_BY = "last_review_by"
    LAST_REVIEW_DATE = "last_review_date"
    LAST_SIGNATURE_BY = "last_signature_by"
    LAST_SIGNATURE_DATE = "last_signature_date"


def _ensure_resource(resource):
    if not set(resource.keys()).issuperset({"bucket", "collection"}):
        msg = "Resources should contain both bucket and collection"
        raise ValueError(msg)
    return resource


class LocalUpdater(object):
    """Sign items in the source and push them to the destination.

    Triggers a signature of all records in the source destination, and
    eventually update the destination with the new signature and the updated
    records.

    :param source:
        Python dictionary containing the bucket and collection of the source.

    :param destination:
        Python dictionary containing the bucket and collection of the
        destination.

    :param signer:
        The instance of the signer that will be used to generate the signature
        on the collection.

    :param storage:
        The instance of kinto.core.storage that will be used to retrieve
        records from the source and add new items to the destination.
    """

    def __init__(self, source, destination, signer, storage, permission):
        self._source = None
        self._destination = None

        self.source = source
        self.destination = destination
        self.signer = signer
        self.storage = storage
        self.permission = permission

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        self._source = _ensure_resource(source)
        self.source_bucket_uri = "/buckets/%s" % source["bucket"]
        self.source_collection_uri = "/buckets/%s/collections/%s" % (
            source["bucket"],
            source["collection"],
        )

    @property
    def destination(self):
        return self._destination

    @destination.setter
    def destination(self, destination):
        self._destination = _ensure_resource(destination)
        self.destination_bucket_uri = "/buckets/%s" % (self.destination["bucket"])
        self.destination_collection_uri = "/buckets/%s/collections/%s" % (
            self.destination["bucket"],
            self.destination["collection"],
        )

    def sign_and_update_destination(
        self,
        request,
        source_attributes,
        next_source_status=STATUS.SIGNED,
        previous_source_status=None,
        push_records=True,
    ):
        """Sign the specified collection.

        0. Create the destination bucket / collection
        1. Get all the records of the collection
        2. Send all records since the last_modified of the destination
        3. Compute a hash of these records
        4. Ask the signer for a signature
        5. Send the signature to the destination.
        """
        changes_count = 0

        self.create_destination(request)

        if push_records:
            changes_count = self.push_records_to_destination(request)

        records, timestamp = self.get_destination_records(empty_none=False)
        serialized_records = canonical_json(records, timestamp)
        logger.debug(f"{self.source_collection_uri}:\t'{serialized_records}'")
        signature = self.signer.sign(serialized_records)

        self.set_destination_signature(signature, source_attributes, request)
        if next_source_status is not None:
            self.update_source_status(
                next_source_status, request, previous_source_status
            )

        return changes_count

    def refresh_signature(self, request, next_source_status=None):
        """Refresh the signature without moving records."""
        records, timestamp = self.get_destination_records(empty_none=False)
        serialized_records = canonical_json(records, timestamp)
        logger.debug(f"{self.source_collection_uri}:\t'{serialized_records}'")
        signature = self.signer.sign(serialized_records)
        self.set_destination_signature(signature, request=request, source_attributes={})

        if next_source_status is not None:
            current_userid = request.prefixed_userid
            current_date = datetime.datetime.now(datetime.timezone.utc).isoformat()
            attrs = {"status": next_source_status}
            attrs[TRACKING_FIELDS.LAST_SIGNATURE_BY.value] = current_userid
            attrs[TRACKING_FIELDS.LAST_SIGNATURE_DATE.value] = current_date
            self._update_source_attributes(request, **attrs)

    def rollback_changes(
        self, request, refresh_last_edit=True, refresh_signature=False
    ):
        """Restore the contents of *destination* to *source* (delete extras, recreate deleted,
        and restore changes) (eg. destination -> preview, or preview -> source).
        """
        dest_records, _ = self.get_destination_records(empty_none=False)
        dest_by_id = {r["id"]: r for r in dest_records}
        source_records, _ = self.get_source_records()

        changes_since_approval = records_diff(source_records, dest_records)

        storage_kwargs = {
            "parent_id": self.source_collection_uri,
            "resource_name": "record",
        }

        changed_count = 0
        for record in changes_since_approval:
            action = None
            record_before = None
            impacted = None

            dest_record = dest_by_id.get(record[FIELD_ID])
            if dest_record is None:
                # In source, but not in destination. Must be deleted.
                if not record.get("deleted"):
                    tombstone = self.storage.delete(
                        object_id=record[FIELD_ID],
                        last_modified=record[FIELD_LAST_MODIFIED],
                        **storage_kwargs,
                    )
                    action = ACTIONS.DELETE
                    record_before = record
                    impacted = tombstone

            # In dest_records, but not in source_records. Must be re-created.
            elif record.get("deleted"):
                self.storage.create(obj=dest_record, **storage_kwargs)
                action = ACTIONS.CREATE
                record_before = None
                impacted = dest_record

            # Differ, restore attributes of dest_record in source.
            else:
                self.storage.update(
                    object_id=record[FIELD_ID], obj=dest_record, **storage_kwargs
                )
                action = ACTIONS.UPDATE
                record_before = record
                impacted = dest_record

            if action is not None:
                changed_count += 1
                # Notify resource event, in order to leave a trace in the history.
                matchdict = {
                    "bucket_id": self.destination["bucket"],
                    "collection_id": self.destination["collection"],
                    FIELD_ID: record[FIELD_ID],
                }
                record_uri = (
                    "/buckets/{bucket_id}/collections/{collection_id}/records/{id}"
                ).format(**matchdict)

                notify_resource_event(
                    request,
                    {
                        "method": "DELETE" if action == ACTIONS.DELETE else "PUT",
                        "path": record_uri,
                    },
                    matchdict=matchdict,
                    resource_name="record",
                    parent_id=self.source_collection_uri,
                    obj=impacted,
                    action=action,
                    old=record_before,
                )

        if refresh_last_edit:
            current_userid = request.prefixed_userid
            current_date = datetime.datetime.now(datetime.timezone.utc).isoformat()
            attrs = {
                "status": STATUS.SIGNED.value,
                "last_editor_comment": "",
                "last_reviewer_comment": "",
            }
            attrs[TRACKING_FIELDS.LAST_EDIT_BY.value] = current_userid
            attrs[TRACKING_FIELDS.LAST_EDIT_DATE.value] = current_date
            self._update_source_attributes(request, **attrs)

        return changed_count

    def create_destination(self, request):
        """Create the destination bucket/collection if they don't already exist."""
        # With the current implementation, the destination is not writable by
        # anyone and readable by everyone.
        # https://github.com/Kinto/kinto-signer/issues/55
        bucket_name = self.destination["bucket"]
        collection_name = self.destination["collection"]

        # Destination bucket will be writable by current user.
        perms = {"write": [request.prefixed_userid]}
        ensure_resource_exists(
            request=request,
            resource_name="bucket",
            parent_id="",
            obj={FIELD_ID: bucket_name},
            permissions=perms,
            matchdict={"id": bucket_name},
        )

        # Destination collection will be publicly readable.
        readonly_perms = {"read": (Everyone,)}
        ensure_resource_exists(
            request=request,
            resource_name="collection",
            parent_id=self.destination_bucket_uri,
            obj={FIELD_ID: collection_name},
            permissions=readonly_perms,
            matchdict={"bucket_id": bucket_name, "id": collection_name},
        )

    def _get_records(self, resource, empty_none=True):
        bid = resource["bucket"]
        cid = resource["collection"]
        parent_id = f"/buckets/{bid}/collections/{cid}"

        records = self.storage.list_all(parent_id=parent_id, resource_name="record")

        if len(records) == 0 and empty_none:
            # When the collection empty (no records and no tombstones)
            collection_timestamp = None
        else:
            collection_timestamp = self.storage.resource_timestamp(
                parent_id=parent_id, resource_name="record"
            )

        return records, collection_timestamp

    def get_source_records(self, **kwargs):
        return self._get_records(self.source, **kwargs)

    def get_destination_records(self, **kwargs):
        return self._get_records(self.destination, **kwargs)

    def push_records_to_destination(self, request):
        dest_records, dest_timestamp = self.get_destination_records()
        source_records, source_timestamp = self.get_source_records()
        new_records = records_diff(source_records, dest_records)
        changes_count = len(new_records)

        if len(new_records) == 0:
            return

        # Update the destination collection.
        for record in new_records:
            storage_kwargs = {
                "parent_id": self.destination_collection_uri,
                "resource_name": "record",
            }
            try:
                before = self.storage.get(object_id=record[FIELD_ID], **storage_kwargs)
            except RecordNotFoundError:
                before = None

            # Timestamp should be bumped in destination.
            record = {**record}
            del record[FIELD_LAST_MODIFIED]

            deleted = record.get("deleted", False)
            if deleted:
                try:
                    pushed = self.storage.delete(
                        object_id=record[FIELD_ID],
                        **storage_kwargs,
                    )
                    action = ACTIONS.DELETE
                except RecordNotFoundError:
                    # If the record doesn't exists in the destination
                    # we are good and can ignore it.
                    continue
            else:
                if before is None:
                    pushed = self.storage.create(obj=record, **storage_kwargs)
                    action = ACTIONS.CREATE
                else:
                    pushed = self.storage.update(
                        object_id=record[FIELD_ID], obj=record, **storage_kwargs
                    )
                    action = ACTIONS.UPDATE

            matchdict = {
                "bucket_id": self.destination["bucket"],
                "collection_id": self.destination["collection"],
                FIELD_ID: record[FIELD_ID],
            }
            bid = matchdict["bucket_id"]
            cid = matchdict["collection_id"]
            rid = matchdict["id"]
            record_uri = f"/buckets/{bid}/collections/{cid}/records/{rid}"

            notify_resource_event(
                request,
                {"method": "DELETE" if deleted else "PUT", "path": record_uri},
                matchdict=matchdict,
                resource_name="record",
                parent_id=self.destination_collection_uri,
                obj=pushed,
                action=action,
                old=before,
            )

        return changes_count

    def set_destination_signature(self, signature, source_attributes, request):
        # Push the new signature to the destination collection.
        parent_id = "/buckets/%s" % self.destination["bucket"]
        collection_id = "collection"

        collection_record = self.storage.get(
            parent_id=parent_id,
            resource_name=collection_id,
            object_id=self.destination["collection"],
        )

        # Update the collection_record
        new_collection = dict(**collection_record)
        new_collection.pop(FIELD_LAST_MODIFIED, None)
        new_collection["signature"] = signature
        for attr in PUBLISHED_COLLECTION_FIELDS:
            if attr in source_attributes:
                new_collection.setdefault(attr, source_attributes[attr])

        updated = self.storage.update(
            parent_id=parent_id,
            resource_name=collection_id,
            object_id=self.destination["collection"],
            obj=new_collection,
        )

        matchdict = dict(
            bucket_id=self.destination["bucket"], id=self.destination["collection"]
        )
        notify_resource_event(
            request,
            {"method": "PUT", "path": self.destination_collection_uri},
            matchdict=matchdict,
            resource_name="collection",
            parent_id=self.destination_bucket_uri,
            obj=updated,
            action=ACTIONS.UPDATE,
            old=collection_record,
        )

    def update_source_review_request_by(self, request):
        current_date = datetime.datetime.now(datetime.timezone.utc).isoformat()
        attrs = {
            TRACKING_FIELDS.LAST_REVIEW_REQUEST_BY.value: request.prefixed_userid,
            TRACKING_FIELDS.LAST_REVIEW_REQUEST_DATE.value: current_date,
        }
        return self._update_source_attributes(request, **attrs)

    def update_source_status(self, status, request, old_status=None):
        current_userid = request.prefixed_userid
        current_date = datetime.datetime.now(datetime.timezone.utc).isoformat()
        attrs = {"status": status.value}
        if status == STATUS.WORK_IN_PROGRESS:
            attrs[TRACKING_FIELDS.LAST_EDIT_BY.value] = current_userid
            attrs[TRACKING_FIELDS.LAST_EDIT_DATE.value] = current_date
        if status == STATUS.TO_REVIEW:
            attrs[TRACKING_FIELDS.LAST_REVIEW_REQUEST_BY.value] = current_userid
            attrs[TRACKING_FIELDS.LAST_REVIEW_REQUEST_DATE.value] = current_date
            # Make sure we reset the review comments if none is specified.
            if "last_editor_comment" not in request.validated["body"].get("data", {}):
                attrs["last_editor_comment"] = ""
        if status == STATUS.SIGNED:
            if old_status != STATUS.SIGNED:
                # Do not keep track of reviewer when refreshing signature.
                attrs[TRACKING_FIELDS.LAST_REVIEW_BY.value] = current_userid
                attrs[TRACKING_FIELDS.LAST_REVIEW_DATE.value] = current_date
            attrs[TRACKING_FIELDS.LAST_SIGNATURE_BY.value] = current_userid
            attrs[TRACKING_FIELDS.LAST_SIGNATURE_DATE.value] = current_date
        return self._update_source_attributes(request, **attrs)

    def _update_source_attributes(self, request, **kwargs):
        parent_id = "/buckets/%s" % self.source["bucket"]
        resource_name = "collection"

        collection_record = self.storage.get(
            parent_id=parent_id,
            resource_name=resource_name,
            object_id=self.source["collection"],
        )

        # Update the collection_record
        new_collection = dict(**collection_record)
        new_collection.update(**kwargs)

        # Remove last_modified to be sure it's bumped.
        new_collection.pop("last_modified", None)

        updated = self.storage.update(
            parent_id=parent_id,
            resource_name=resource_name,
            object_id=self.source["collection"],
            obj=new_collection,
        )

        matchdict = dict(bucket_id=self.source["bucket"], id=self.source["collection"])
        notify_resource_event(
            request,
            {"method": "PUT", "path": self.source_collection_uri},
            matchdict=matchdict,
            resource_name="collection",
            parent_id=self.source_bucket_uri,
            obj=updated,
            action=ACTIONS.UPDATE,
            old=collection_record,
        )



import codecs
import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))


def read_file(filename):
    """Open a related file and return its content."""
    with codecs.open(os.path.join(here, filename), encoding="utf-8") as f:
        content = f.read()
    return content


README = read_file("README.rst")
CHANGELOG = read_file("CHANGELOG.rst")

INSTALL_REQUIRES = [
    x.replace(" \\", "")
    for x in read_file("./requirements.txt").split("\n")
    if not x.startswith(" ")
]

setup(
    name="kinto-dist",
    version="27.0.3",
    description="Kinto Distribution",
    long_description=README + "\n\n" + CHANGELOG,
    license="Apache License (2.0)",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "License :: OSI Approved :: Apache Software License",
    ],
    keywords="web services",
    author="Mozilla Services",
    author_email="services-dev@mozilla.com",
    url="https://github.com/mozilla-services/kinto-dist",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=INSTALL_REQUIRES,
)



import os.path
import time
import unittest
from urllib.parse import urljoin

import requests
from kinto_http import Client, KintoException
from kinto_http.patch_type import JSONPatch

from kinto_remote_settings.signer.backends import local_ecdsa
from kinto_remote_settings.signer.serializer import canonical_json

__HERE__ = os.path.abspath(os.path.dirname(__file__))

SERVER_URL = "http://localhost:8888/v1"
DEFAULT_AUTH = ("user", "p4ssw0rd")


def collection_timestamp(client, **kwargs):
    # XXXX Waiting https://github.com/Kinto/kinto-http.py/issues/77
    endpoint = client.get_endpoint("records", **kwargs)
    record_resp, headers = client.session.request("get", endpoint)
    return headers.get("ETag", "").strip('"')


def user_principal(client):
    return client.server_info()["user"]["id"]


def create_records(client):
    # Create some data on the client collection and send it.
    with client.batch() as batch:
        for n in range(0, 10):
            batch.create_record(data={"newdata": n})


def flush_server(server_url):
    flush_url = urljoin(server_url, "/v1/__flush__")
    resp = requests.post(flush_url)
    resp.raise_for_status()


def trigger_signature(editor_client, reviewer_client=None):
    editor_client.patch_collection(data={"status": "to-review"})
    reviewer_client.patch_collection(data={"status": "to-sign"})


def fetch_history(client):
    url = client.get_endpoint("bucket") + "/history"
    body, headers = client.session.request("GET", url)
    return body["data"]


def create_group(client, name, members):
    endpoint = client.get_endpoint("collections")
    endpoint = endpoint.replace("/collections", "/groups/%s" % name)
    data = {"members": members}
    resp, headers = client.session.request("put", endpoint, data)
    return resp


class BaseTestFunctional(object):
    server_url = SERVER_URL

    @classmethod
    def setUpClass(cls):
        super(BaseTestFunctional, cls).setUpClass()
        cls.signer = local_ecdsa.ECDSASigner(private_key=cls.private_key)
        cls.source = Client(
            server_url=cls.server_url,
            auth=DEFAULT_AUTH,
            bucket=cls.source_bucket,
            collection=cls.source_collection,
        )
        cls.destination = Client(
            server_url=cls.server_url,
            auth=DEFAULT_AUTH,
            bucket=cls.destination_bucket,
            collection=cls.destination_collection,
        )
        cls.editor_client = Client(
            server_url=cls.server_url,
            auth=("editor", ""),
            bucket=cls.source_bucket,
            collection=cls.source_collection,
        )
        cls.someone_client = Client(
            server_url=cls.server_url,
            auth=("Sam", "Wan-Elss"),
            bucket=cls.source_bucket,
            collection=cls.source_collection,
        )

    @classmethod
    def tearDown(cls):
        # Delete all the created objects.
        flush_server(cls.server_url)

    def setUp(self):
        # Give the permission to write in collection to anybody
        self.source.create_bucket()
        perms = {"write": ["system.Authenticated"]}
        self.source.create_collection(permissions=perms)
        principals = [
            user_principal(self.editor_client),
            user_principal(self.someone_client),
            user_principal(self.source),
        ]
        create_group(self.source, "editors", members=principals)
        create_group(self.source, "reviewers", members=principals)

        # Create some data on the source collection and send it.
        create_records(self.source)

        self.source_records = self.source.get_records()
        assert len(self.source_records) == 10

        time.sleep(0.1)

        trigger_signature(editor_client=self.editor_client, reviewer_client=self.source)

    def test_groups_and_reviewers_are_forced(self):
        capability = self.source.server_info()["capabilities"]["signer"]
        assert capability["group_check_enabled"]
        assert capability["to_review_enabled"]

    def test_heartbeat_is_successful(self):
        hb_url = urljoin(self.server_url, "/__heartbeat__")
        resp = requests.get(hb_url)
        resp.raise_for_status()

    def test_metadata_attributes(self):
        # Ensure the destination data is signed properly.
        destination_collection = self.destination.get_collection()["data"]
        signature = destination_collection["signature"]
        assert signature is not None

        # the status of the source collection should be "signed".
        source_collection = self.source.get_collection()["data"]
        assert source_collection["status"] == "signed"

        assert collection_timestamp(self.source) == collection_timestamp(self.source)

    def test_destination_creation_and_new_records_signature(self):
        # Create some records and trigger another signature.
        self.source.create_record(data={"newdata": "hello"})
        self.source.create_record(data={"newdata": "bonjour"})

        time.sleep(0.1)

        trigger_signature(editor_client=self.editor_client, reviewer_client=self.source)
        data = self.destination.get_collection()
        signature = data["data"]["signature"]
        assert signature is not None

        records = self.destination.get_records()
        assert len(records) == 12
        last_modified = collection_timestamp(self.destination)
        serialized_records = canonical_json(records, last_modified)
        # This raises when the signature is invalid.
        self.signer.verify(serialized_records, signature)

    def test_records_update_and_signature(self):
        # Update some records and trigger another signature.
        updated = self.source_records[5].copy()
        updated["newdata"] = "bump"
        self.source.update_record(data=updated)
        updated = self.source_records[0].copy()
        updated["newdata"] = "hoop"
        self.source.update_record(data=updated)

        time.sleep(0.1)

        trigger_signature(editor_client=self.editor_client, reviewer_client=self.source)
        data = self.destination.get_collection()
        signature = data["data"]["signature"]
        assert signature is not None

        records = self.destination.get_records()
        assert len(records) == 10
        last_modified = collection_timestamp(self.destination)
        serialized_records = canonical_json(records, last_modified)
        # This raises when the signature is invalid.
        self.signer.verify(serialized_records, signature)

    def test_records_deletion_and_signature(self):
        # Now delete one record on the source and trigger another signature.
        self.source.delete_record(id=self.source_records[1]["id"])
        self.source.delete_record(id=self.source_records[5]["id"])

        time.sleep(0.1)

        trigger_signature(editor_client=self.editor_client, reviewer_client=self.source)

        data = self.destination.get_collection()
        signature = data["data"]["signature"]
        assert signature is not None

        records = self.destination.get_records(_since=0)  # obtain deleted too
        last_modified = collection_timestamp(self.destination)
        serialized_records = canonical_json(records, last_modified)

        assert len(records) == 10  # two of them are deleted.
        assert len([r for r in records if r.get("deleted", False)]) == 2

        # This raises when the signature is invalid.
        self.signer.verify(serialized_records, signature)

    def test_records_delete_all_and_signature(self):
        source_records = self.source.get_records()
        destination_records = self.destination.get_records()

        assert len(source_records) == len(destination_records)

        self.source.delete_records()

        trigger_signature(editor_client=self.editor_client, reviewer_client=self.source)

        source_records = self.source.get_records()
        destination_records = self.destination.get_records()

        assert len(source_records) == len(destination_records) == 0

        last_modified = collection_timestamp(self.destination)
        serialized_records = canonical_json(destination_records, last_modified)

        data = self.destination.get_collection()
        signature = data["data"]["signature"]
        assert signature is not None

        # This raises when the signature is invalid.
        self.signer.verify(serialized_records, signature)

    def test_distinct_users_can_trigger_signatures(self):
        collection = self.destination.get_collection()
        before = collection["data"]["signature"]

        self.source.create_record(data={"pim": "pam"})
        # Trigger a signature as someone else.
        trigger_signature(
            editor_client=self.editor_client, reviewer_client=self.someone_client
        )

        collection = self.destination.get_collection()
        after = collection["data"]["signature"]

        assert before != after


class AliceFunctionalTest(BaseTestFunctional, unittest.TestCase):
    private_key = os.path.join(__HERE__, "config/ecdsa.private.pem")
    source_bucket = "alice"
    destination_bucket = "alice"
    source_collection = "source"
    destination_collection = "destination"


# Signer is configured to use a different key for Bob and Alice.
class BobFunctionalTest(BaseTestFunctional, unittest.TestCase):
    private_key = os.path.join(__HERE__, "config/bob.ecdsa.private.pem")
    source_bucket = "bob"
    source_collection = "source"
    destination_bucket = "bob"
    destination_collection = "destination"


# Signoff is configured per bucket.
class PerBucketFunctionalTest(BaseTestFunctional, unittest.TestCase):
    private_key = os.path.join(__HERE__, "config/ecdsa.private.pem")
    source_bucket = "stage"
    source_collection = "cid"
    destination_bucket = "prod"
    destination_collection = "cid"


class HistoryTest(unittest.TestCase):
    server_url = SERVER_URL
    private_key = os.path.join(__HERE__, "config/ecdsa.private.pem")
    source_bucket = "alice"
    source_collection = "source"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.signer = local_ecdsa.ECDSASigner(private_key=cls.private_key)
        cls.source = Client(
            server_url=cls.server_url,
            auth=DEFAULT_AUTH,
            bucket=cls.source_bucket,
            collection=cls.source_collection,
        )
        cls.editor_client = Client(
            server_url=cls.server_url,
            auth=("editor", ""),
            bucket=cls.source_bucket,
            collection=cls.source_collection,
        )

    def tearDown(self):
        # Delete all the created objects.
        flush_server(self.server_url)

    def setUp(self):
        # Give the permission to write in collection to anybody
        self.source.create_bucket()
        principals = [user_principal(self.editor_client), user_principal(self.source)]
        create_group(self.source, "editors", members=principals)
        create_group(self.source, "reviewers", members=principals)

        perms = {"write": ["system.Authenticated"]}
        self.source.create_collection(permissions=perms)
        self.source.create_record(data={"hola": "mundo"})
        trigger_signature(editor_client=self.editor_client, reviewer_client=self.source)

    def test_history_entries_contain_signer_actions(self):
        entries = fetch_history(self.source)
        entries.reverse()
        collection_entries = [
            e
            for e in entries
            if e["resource_name"] == "collection" and e["collection_id"] == "source"
        ]
        assert len(collection_entries) == 7

        # create collection
        assert collection_entries[0]["action"] == "create"
        assert "basicauth:" in collection_entries[0]["user_id"]
        assert collection_entries[1]["action"] == "update"
        assert "kinto-signer" in collection_entries[1]["user_id"]

        # status: work-in-progress
        assert collection_entries[2]["target"]["data"]["status"] == "work-in-progress"
        assert "kinto-signer" in collection_entries[2]["user_id"]
        assert collection_entries[4]["target"]["data"]["status"] == "to-review"

        # status: to-review (by user)
        assert collection_entries[3]["target"]["data"]["status"] == "to-review"
        assert "last_review_request_by" not in collection_entries[2]["target"]["data"]
        assert "basicauth:" in collection_entries[3]["user_id"]

        # update of last_editor (by plugin)
        assert collection_entries[4]["target"]["data"]["status"] == "to-review"
        assert (
            "basicauth:"
            in collection_entries[4]["target"]["data"]["last_review_request_by"]
        )
        assert "kinto-signer" in collection_entries[4]["user_id"]

        # status: to-sign
        assert collection_entries[5]["target"]["data"]["status"] == "to-sign"
        assert "last_review_by" not in collection_entries[5]["target"]["data"]
        assert "basicauth:" in collection_entries[5]["user_id"]

        # status: signed (by plugin)
        assert collection_entries[6]["target"]["data"]["status"] == "signed"
        assert "basicauth:" in collection_entries[6]["target"]["data"]["last_review_by"]
        assert "kinto-signer" in collection_entries[6]["user_id"]


class WorkflowTest(unittest.TestCase):
    server_url = SERVER_URL

    @classmethod
    def setUpClass(cls):
        super(WorkflowTest, cls).setUpClass()
        client_kw = dict(server_url=cls.server_url, bucket="alice", collection="from")
        cls.client = Client(auth=DEFAULT_AUTH, **client_kw)
        cls.elsa_client = Client(auth=("elsa", ""), **client_kw)
        cls.anna_client = Client(auth=("anna", ""), **client_kw)
        cls.client_principal = user_principal(cls.client)
        cls.elsa_principal = user_principal(cls.elsa_client)
        cls.anna_principal = user_principal(cls.anna_client)

        private_key = os.path.join(__HERE__, "config/ecdsa.private.pem")
        cls.signer = local_ecdsa.ECDSASigner(private_key=private_key)

    def setUp(self):
        perms = {"write": ["system.Authenticated"]}
        self.client.create_bucket()
        create_group(
            self.client, "editors", members=[self.anna_principal, self.client_principal]
        )
        create_group(
            self.client,
            "reviewers",
            members=[self.elsa_principal, self.client_principal],
        )
        self.client.create_collection(permissions=perms)

    def tearDown(self):
        # Delete all the created objects.
        flush_server(self.server_url)

    def test_status_work_in_progress(self):
        collection = self.client.get_collection()
        assert collection["data"]["status"] == "signed"

        create_records(self.client)

        collection = self.client.get_collection()
        after = collection["data"]["status"]
        assert after == "work-in-progress"

    def test_whole_workflow(self):
        create_records(self.client)
        self.anna_client.patch_collection(data={"status": "to-review"})
        self.elsa_client.patch_collection(data={"status": "to-sign"})
        status = self.client.get_collection()["data"]["status"]
        assert status == "signed"

    def test_only_editors_can_ask_for_review(self):
        with self.assertRaises(KintoException):
            self.elsa_client.patch_collection(data={"status": "to-review"})

    def test_status_can_be_maintained_as_to_review(self):
        self.anna_client.patch_collection(data={"status": "to-review"})
        self.elsa_client.patch_collection(data={"status": "to-review"})

    def test_preview_collection_is_updated_and_signed_on_to_review(self):
        create_records(self.client)
        self.anna_client.patch_collection(data={"status": "to-review"})

        collection = self.client.get_collection(id="preview")
        records = self.client.get_records(collection="preview")
        last_modified = collection_timestamp(self.client, collection="preview")
        serialized_records = canonical_json(records, last_modified)

        signature = collection["data"]["signature"]
        assert signature is not None
        self.signer.verify(serialized_records, signature)

        assert len(records) == 10

    def test_same_editor_cannot_review(self):
        self.anna_client.patch_collection(data={"status": "to-review"})
        with self.assertRaises(KintoException):
            self.anna_client.patch_collection(data={"status": "to-sign"})

    def test_status_cannot_be_set_to_sign_without_review(self):
        create_records(self.client)
        with self.assertRaises(KintoException):
            self.elsa_client.patch_collection(data={"status": "to-sign"})

    def test_changes_can_be_rolledback(self):
        destination_records = self.client.get_records(collection="to")
        assert len(destination_records) == 0
        create_records(self.client)
        source_records = self.client.get_records(collection="from")
        assert len(source_records) == 10
        self.anna_client.patch_collection(data={"status": "to-review"})
        preview_records = self.client.get_records(collection="preview")
        assert len(preview_records) == 10

        self.anna_client.patch_collection(data={"status": "to-rollback"})

        source_records = self.client.get_records(collection="from")
        assert len(source_records) == len(destination_records)
        preview_records = self.client.get_records(collection="preview")
        assert len(preview_records) == len(destination_records)

    def test_review_can_be_cancelled_by_editor(self):
        create_records(self.client)
        self.anna_client.patch_collection(data={"status": "to-review"})
        self.anna_client.patch_collection(data={"status": "work-in-progress"})
        self.anna_client.patch_collection(data={"status": "to-review"})
        self.elsa_client.patch_collection(data={"status": "to-sign"})

    def test_review_can_be_cancelled_by_reviewer(self):
        create_records(self.client)
        self.anna_client.patch_collection(data={"status": "to-review"})
        self.elsa_client.patch_collection(data={"status": "work-in-progress"})
        create_records(self.anna_client)
        self.anna_client.patch_collection(data={"status": "to-review"})
        self.elsa_client.patch_collection(data={"status": "to-sign"})

    def test_must_ask_for_review_after_cancelled(self):
        create_records(self.client)
        self.anna_client.patch_collection(data={"status": "to-review"})
        self.elsa_client.patch_collection(data={"status": "work-in-progress"})
        with self.assertRaises(KintoException):
            self.elsa_client.patch_collection(data={"status": "to-sign"})

    def test_editors_can_be_different_after_cancelled(self):
        create_records(self.client)
        self.client.patch_collection(data={"status": "to-review"})

        resp = self.client.get_collection()
        assert resp["data"]["last_review_request_by"] == self.client_principal

        # Client cannot review since he is the last_editor.
        with self.assertRaises(KintoException):
            self.client.patch_collection(data={"status": "to-sign"})
        # Someone rejects the review.
        self.elsa_client.patch_collection(data={"status": "work-in-progress"})
        # Anna becomes the last_editor.
        self.anna_client.patch_collection(data={"status": "to-review"})
        # Client can now review because he is not the last_editor.
        self.client.patch_collection(data={"status": "to-sign"})

    def test_modifying_the_collection_resets_status(self):
        create_records(self.client)
        self.anna_client.patch_collection(data={"status": "to-review"})
        create_records(self.client)
        status = self.client.get_collection()["data"]["status"]
        assert status == "work-in-progress"

    def test_can_refresh_if_never_signed(self):
        create_records(self.elsa_client)
        source_data = self.client.get_collection()["data"]
        assert source_data["status"] == "work-in-progress"
        destination_data = self.client.get_collection(id="to")["data"]
        before_signature = destination_data.get("signature")

        self.elsa_client.patch_collection(data={"status": "to-resign"})

        source_data = self.client.get_collection()["data"]
        assert source_data["status"] == "work-in-progress"
        assert "last_review_request_date" not in source_data
        assert "last_signature_date" in source_data
        destination_data = self.client.get_collection(id="to")["data"]
        assert destination_data["signature"] != before_signature
        # Refresh does not copy records.
        destination_records = self.client.get_records(collection="to")
        assert len(destination_records) == 0

    def test_refresh_signs_preview_collection(self):
        preview_data = self.client.get_collection(id="preview")["data"]
        before_signature = preview_data.get("signature")

        self.elsa_client.patch_collection(data={"status": "to-resign"})

        preview_data = self.client.get_collection(id="preview")["data"]
        assert preview_data["signature"] != before_signature


class PerBucketTest(unittest.TestCase):
    server_url = SERVER_URL

    @classmethod
    def setUpClass(cls):
        super(PerBucketTest, cls).setUpClass()
        client_kw = dict(server_url=cls.server_url, bucket="stage")
        cls.client = Client(auth=DEFAULT_AUTH, **client_kw)
        cls.anon_client = Client(auth=tuple(), **client_kw)
        cls.julia_client = Client(auth=("julia", ""), **client_kw)
        cls.joan_client = Client(auth=("joan", ""), **client_kw)
        cls.julia_principal = user_principal(cls.julia_client)
        cls.joan_principal = user_principal(cls.joan_client)

    def setUp(self):
        self.client.create_bucket(
            permissions={
                "collection:create": ["system.Authenticated"],
                "group:create": ["system.Authenticated"],
            }
        )

    def tearDown(cls):
        # Delete all the created objects.
        flush_server(cls.server_url)

    def test_anyone_can_create_collections(self):
        collection = self.julia_client.create_collection(id="pim")
        assert self.julia_principal in collection["permissions"]["write"]

    def test_editors_and_reviewers_groups_are_created(self):
        self.julia_client.create_collection(id="pam")
        editors_group = self.julia_client.get_group(id="editors")
        reviewers_group = self.julia_client.get_group(id="reviewers")
        assert self.julia_principal in editors_group["data"]["members"]
        assert self.julia_principal not in reviewers_group["data"]["members"]

    def test_preview_and_destination_collections_are_signed(self):
        self.julia_client.create_collection(id="poum")

        preview_collection = self.anon_client.get_collection(
            bucket="preview", id="poum"
        )
        assert "signature" in preview_collection["data"]

        prod_collection = self.anon_client.get_collection(bucket="prod", id="poum")
        assert "signature" in prod_collection["data"]

    def test_collection_can_be_deleted(self):
        self.julia_client.create_collection(id="poum")

        self.julia_client.delete_collection(id="poum")

        # The following objects are still here, thus not raising:
        self.anon_client.get_collection(bucket="preview", id="poum")
        self.anon_client.get_collection(bucket="prod", id="poum")
        self.julia_client.get_group(id="editors")
        self.julia_client.get_group(id="reviewers")

    def test_full_review_test(self):
        # Create a collection.
        self.julia_client.create_collection(id="pim")

        # Add Joan to reviewers.
        data = JSONPatch(
            [{"op": "add", "path": "/data/members/0", "value": self.joan_principal}]
        )
        self.julia_client.patch_group(id="reviewers", changes=data)

        # Create some records.
        self.julia_client.create_record(id="abc", collection="pim")
        record = self.julia_client.create_record(id="def", collection="pim")
        timestamp_before_approval = record["data"]["last_modified"]

        # Preview and prod have no records yet.
        records = self.anon_client.get_records(bucket="preview", collection="pim")
        assert len(records) == 0
        records = self.anon_client.get_records(bucket="prod", collection="pim")
        assert len(records) == 0

        # Ask for review.
        self.julia_client.patch_collection(id="pim", data={"status": "to-review"})

        # Preview now has records.
        records = self.anon_client.get_records(bucket="preview", collection="pim")
        assert len(records) == 2

        # Approve changes.
        self.joan_client.patch_collection(id="pim", data={"status": "to-sign"})

        # Prod now has records.
        records = self.anon_client.get_records(bucket="prod", collection="pim")
        assert len(records) == 2
        # Publishing to destination bumped the timestamps.
        assert records[0]["last_modified"] != timestamp_before_approval

        # Refresh signature
        self.joan_client.patch_collection(id="pim", data={"status": "to-resign"})

        # Delete source collection.
        self.julia_client.delete_collection(id="pim")
        records = self.anon_client.get_records(bucket="preview", collection="pim")
        assert len(records) == 0
        records = self.anon_client.get_records(bucket="prod", collection="pim")
        assert len(records) == 0



import operator

import canonicaljson


def canonical_json(records, last_modified):
    records = (r for r in records if not r.get("deleted", False))
    records = sorted(records, key=operator.itemgetter("id"))

    payload = {"data": records, "last_modified": "%s" % last_modified}

    dump = canonicaljson.dumps(payload)

    return dump



import unittest
from unittest import mock

from pyramid.request import Request

from kinto_remote_settings.changes.utils import changes_object


class ChangesRecordTest(unittest.TestCase):
    def test_single_hardcoded(self):
        request = Request.blank(path="/")
        request.route_path = mock.Mock()
        request.route_path.return_value = "/buckets/a/collections/b"
        request.registry = mock.Mock()
        request.registry.settings = {}
        timestamp = 1525457597166
        entry = changes_object(request, "a", "b", timestamp)

        self.assertEqual(
            entry,
            {
                "bucket": "a",
                "collection": "b",
                "host": "",
                "id": "9527d115-6191-fa49-a530-8fbfc4997755",
                "last_modified": timestamp,
            },
        )

    def test_another_hardcoded(self):
        request = Request.blank(path="/")
        request.route_path = mock.Mock()
        request.route_path.return_value = "/buckets/a/collections/b"
        request.registry = mock.Mock()
        request.registry.settings = {"http_host": "https://localhost:443"}
        timestamp = 1525457597166
        entry = changes_object(request, "a", "b", timestamp)

        self.assertEqual(
            entry,
            {
                "bucket": "a",
                "collection": "b",
                "host": "https://localhost:443",
                "id": "fa48a96d-1600-f561-8645-3395acb08a5a",
                "last_modified": timestamp,
            },
        )



import os

try:
    import ConfigParser as configparser
except ImportError:
    import configparser

from kinto import main as kinto_main
from kinto.core.testing import BaseWebTest as CoreWebTest
from kinto.core.testing import DummyRequest, get_user_headers

__all__ = ["BaseWebTest", "DummyRequest", "get_user_headers"]


here = os.path.abspath(os.path.dirname(__file__))


class BaseWebTest(CoreWebTest):
    api_prefix = "v1"
    entry_point = kinto_main
    config = "config/signer.ini"

    def __init__(self, *args, **kwargs):
        super(BaseWebTest, self).__init__(*args, **kwargs)
        self.headers.update(get_user_headers("mat"))

    @classmethod
    def get_app_settings(cls, extras=None):
        ini_path = os.path.join(here, cls.config)
        config = configparser.ConfigParser()
        config.read(ini_path)
        settings = dict(config.items("app:main"))
        settings["signer.to_review_enabled"] = False
        return settings



import datetime
import unittest
from unittest import mock

from kinto_remote_settings import __version__

from . import BaseWebTest

SAMPLE_RECORD = {"data": {"dev-edition": True}}
HOUR_AGO = int(datetime.datetime.now().timestamp() * 1000) - 3600


class UpdateChangesTest(BaseWebTest, unittest.TestCase):
    changes_uri = "/buckets/monitor/collections/changes/records"
    records_uri = "/buckets/blocklists/collections/certificates/records"

    def setUp(self):
        super(UpdateChangesTest, self).setUp()
        self.app.post_json(self.records_uri, SAMPLE_RECORD, headers=self.headers)

    def test_parent_bucket_and_collection_dont_have_to_exist(self):
        self.app.delete(
            "/buckets/monitor/collections/changes",
            headers=self.headers,
            status=(403, 404),
        )
        self.app.get(self.changes_uri)  # Not failing
        self.app.delete("/buckets/monitor", headers=self.headers, status=(403, 404))
        self.app.get(self.changes_uri)  # Not failing

    def test_parent_bucket_and_collection_can_exist(self):
        self.app.put("/buckets/monitor", headers=self.headers)
        resp = self.app.get(self.changes_uri)  # Not failing
        self.assertEqual(len(resp.json["data"]), 1)

        self.app.put("/buckets/monitor/collections/changes", headers=self.headers)
        resp = self.app.get(self.changes_uri)  # Not failing
        self.assertEqual(len(resp.json["data"]), 1)

    def test_a_change_record_is_updated_per_bucket_collection(self):
        resp = self.app.get(self.changes_uri)
        before_timestamp = resp.json["data"][0]["last_modified"]
        before_id = resp.json["data"][0]["id"]

        self.app.post_json(self.records_uri, SAMPLE_RECORD, headers=self.headers)

        resp = self.app.get(self.changes_uri)

        after_timestamp = resp.json["data"][0]["last_modified"]
        after_id = resp.json["data"][0]["id"]
        self.assertEqual(before_id, after_id)
        self.assertNotEqual(before_timestamp, after_timestamp)

    def test_only_collections_specified_in_settings_are_monitored(self):
        resp = self.app.get(self.changes_uri, headers=self.headers)
        change_record = resp.json["data"][0]
        records_uri = "/buckets/default/collections/certificates/records"

        self.app.post_json(records_uri, SAMPLE_RECORD, headers=self.headers)

        resp = self.app.get(self.changes_uri, headers=self.headers)
        after = resp.json["data"][0]
        self.assertEqual(change_record["id"], after["id"])
        self.assertEqual(change_record["last_modified"], after["last_modified"])

    def test_the_resource_configured_can_be_a_collection_uri(self):
        with mock.patch.dict(
            self.app.app.registry.settings,
            [("changes.resources", "/buckets/blocklists/collections/certificates")],
        ):
            resp = self.app.get(self.changes_uri)
        self.assertEqual(len(resp.json["data"]), 1)

    def test_returns_304_if_no_change_occured(self):
        resp = self.app.get(self.changes_uri)
        before_timestamp = resp.headers["ETag"]
        self.app.get(
            self.changes_uri, headers={"If-None-Match": before_timestamp}, status=304
        )

    def test_returns_412_with_if_none_match_star(self):
        self.app.get(self.changes_uri, headers={"If-None-Match": "*"}, status=412)

    def test_no_cache_control_is_returned_if_not_configured(self):
        resp = self.app.get(self.changes_uri)
        assert "max-age" not in resp.headers["Cache-Control"]

        resp = self.app.get(self.changes_uri + '?_expected="42"')
        assert "max-age" not in resp.headers["Cache-Control"]

    def test_returns_empty_list_if_no_resource_configured(self):
        with mock.patch.dict(
            self.app.app.registry.settings, [("changes.resources", "")]
        ):
            resp = self.app.get(self.changes_uri)
        self.assertEqual(resp.json["data"], [])

    def test_change_record_has_greater_last_modified_of_collection_of_records(self):
        resp = self.app.post_json(self.records_uri, SAMPLE_RECORD, headers=self.headers)
        last_modified = resp.json["data"]["last_modified"]
        resp = self.app.get(self.changes_uri, headers=self.headers)
        change_last_modified = resp.json["data"][0]["last_modified"]
        self.assertGreaterEqual(change_last_modified, last_modified)

    def test_record_with_old_timestamp_does_update_changes(self):
        resp = self.app.post_json(self.records_uri, SAMPLE_RECORD, headers=self.headers)
        old_record = SAMPLE_RECORD.copy()
        old_record["data"]["last_modified"] = 42
        self.app.post_json(self.records_uri, old_record, headers=self.headers)

        resp = self.app.get(self.changes_uri, headers=self.headers)
        change_last_modified = resp.json["data"][0]["last_modified"]
        self.assertNotEqual(change_last_modified, 42)

    def test_change_record_has_server_host_attribute(self):
        self.app.post_json(self.records_uri, SAMPLE_RECORD, headers=self.headers)

        resp = self.app.get(self.changes_uri, headers=self.headers)
        change = resp.json["data"][0]
        self.assertEqual(change["host"], "www.kinto-storage.org")

    def test_change_record_has_bucket_and_collection_attributes(self):
        self.app.post_json(self.records_uri, SAMPLE_RECORD, headers=self.headers)

        resp = self.app.get(self.changes_uri, headers=self.headers)
        change = resp.json["data"][0]
        self.assertEqual(change["bucket"], "blocklists")
        self.assertEqual(change["collection"], "certificates")

    def test_changes_capability_exposed(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("changes", capabilities)
        expected = {
            "description": "Track modifications of records in Kinto and store "
            "the collection timestamps into a specific bucket "
            "and collection.",
            "collections": ["/buckets/blocklists"],
            "url": "http://kinto.readthedocs.io/en/latest/tutorials/"
            "synchronisation.html#polling-for-remote-changes",
            "version": __version__,
        }
        self.assertEqual(expected, capabilities["changes"])


class CacheExpiresTest(BaseWebTest, unittest.TestCase):
    changes_uri = "/buckets/monitor/collections/changes/records"

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["monitor.changes.record_cache_expires_seconds"] = "60"
        settings["monitor.changes.record_cache_maximum_expires_seconds"] = "3600"
        return settings

    def test_cache_expires_headers_are_supported(self):
        resp = self.app.get(self.changes_uri)
        assert "max-age=60" in resp.headers["Cache-Control"]

    def test_cache_expires_header_is_maximum_with_cache_busting(self):
        resp = self.app.get(self.changes_uri + f"?_since={HOUR_AGO}&_expected=42")
        assert "max-age=3600" in resp.headers["Cache-Control"]

    def test_cache_expires_header_is_default_with_filter(self):
        # The _since just filters on lower bound of timestamps, if data changes
        # we don't want to cache for too long.
        resp = self.app.get(self.changes_uri + f"?_since={HOUR_AGO}")
        assert "max-age=60" in resp.headers["Cache-Control"]

    def test_cache_expires_header_is_default_with_concurrency_control(self):
        # The `If-None-Match` header is just a way to obtain a 304 instead of a 200
        # with an empty list. In the client code [0] it is always used in conjonction
        # with _since={last-etag}
        # [0] https://searchfox.org/mozilla-central/rev/93905b66/services/settings/Utils.jsm#70-73 # noqa: 501
        headers = {"If-None-Match": f'"{HOUR_AGO}"'}
        resp = self.app.get(self.changes_uri + f'?_since="{HOUR_AGO}"', headers=headers)
        assert "max-age=60" in resp.headers["Cache-Control"]


class OldSinceRedirectTest(BaseWebTest, unittest.TestCase):
    changes_uri = "/buckets/monitor/collections/changes/records"

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["kinto.changes.since_max_age_days"] = "2"
        settings["kinto.changes.http_host"] = "cdn-host"
        return settings

    def test_redirects_and_drops_since_if_too_old(self):
        resp = self.app.get(self.changes_uri + "?_since=42")
        self.assertEqual(resp.status_code, 307)
        self.assertEqual(
            resp.headers["Location"],
            "https://cdn-host/v1/buckets/monitor/collections/changes/records",
        )

        # Try again with a real timestamp older than allowed in settings.
        timestamp = int(
            (datetime.datetime.now() - datetime.timedelta(days=3)).timestamp() * 1000
        )
        resp = self.app.get(self.changes_uri + f"?_since={timestamp}")
        self.assertEqual(resp.status_code, 307)

    def test_redirects_keep_other_querystring_params(self):
        resp = self.app.get(self.changes_uri + "?_since=42&_expected=%22123456%22")
        self.assertEqual(resp.status_code, 307)
        self.assertIn("/records?_expected=%22123456%22", resp.headers["Location"])

    def test_does_not_redirect_if_not_old_enough(self):
        timestamp = int(
            (datetime.datetime.now() - datetime.timedelta(days=1)).timestamp() * 1000
        )
        resp = self.app.get(self.changes_uri + f"?_since={timestamp}")
        self.assertEqual(resp.status_code, 200)

    def test_redirects_sends_cache_control(self):
        response = self.app.get(self.changes_uri + "?_since=42")
        self.assertEqual(response.status_code, 307)
        self.assertIn("Expires", response.headers)
        self.assertIn("Cache-Control", response.headers)
        self.assertEqual(response.headers["Cache-Control"], "max-age=86400")



from kinto import logger


def heartbeat(request):
    """Test that signer is operationnal.

    :param request: current request object
    :type request: :class:`~pyramid:pyramid.request.Request`
    :returns: ``True`` is everything is ok, ``False`` otherwise.
    :rtype: bool
    """
    for signer in request.registry.signers.values():
        try:
            signer.sign("This is a heartbeat test.")
        except Exception as e:
            logger.exception(e)
            return False
    return True
